

local BASE, VERSION, MINOR = 'LibBuilder', '1.0', 2
local MAJOR = BASE..'-'..VERSION


local core = LibStub:NewLibrary(MAJOR, MINOR)
if not core then return end


--	external upvalues
local _G = _G
local LibStub = _G.LibStub
local type = _G.type
local select = _G.select


local shellDefaultVersion = '1.0'	--	used if we supply the shell

--	initialize fields
--	leave clientMajor, clientMinor, and clientWorkspace intact
--	Aliases are 'virtual' methods that are only 
--	available through __call and are not recursive.
core.aliases = {
	[''] = 'GetLib', 	--	shortcut for accessing through shellcall
	new = 'NewLib',
	get = 'GetLib',
	close = 'CloseLib',
}
core.errorStrings = {
	TEMPLATE = MAJOR..': %s!', 
	BADVERSION = '%s requires a textual major and numeric minor version', 
	NOLIB = '%s called but no library is being built', 
	LEFTOPEN = '%s, Rev. %s was not closed properly',
	SHELLNODEFAULT = 'The default version is unavailable',
}

--	reuse the old frame if it exists
if core.frame == nil then
	--	Create a frame so we can throw a warning if 
	--	a library is still open when OnUpdate fires.
	--	Use OnUpdate instead of ADDON_LOADED because an error in the client 
	--	library would load Blizzard_DebugTools, confuse the builder, and 
	--	seemed to prevent the real error message from showing.
	local frame = CreateFrame('Frame')
	core.frame = frame
	
	frame:Hide()	--	we don't want to hide it if we're upgrading because a client lib may be open
	frame:SetScript('OnUpdate', function() core:OnUpdate() end)
end


--	Internal; uses frame updates to check if a library was left open.
function core:OnUpdate()
	--	hide frame so we don't get another update unless another library 
	--	is created, and do it now in case we have to throw the error
	self.frame:Hide()

	if self.clientMajor then
		--	Close it so the error only happens once.
		--	We have to cache the version info because it'll get wiped.
		local major, minor = self:CloseLib()
		
		local message = self:BuildErrorMessage('LEFTOPEN', major, minor)
		error(message)
	end
end


---	Creates a workspace for a new client library.
--	@name LibBuilder:NewLib([register,] major, minor [, workspace])
--	@param		register	Optional, defaults to true. If true, this function will attempt to create the library in LibStub.
--	@param		major		The major version of the client library.
--	@param		minor		The minor version of the client library.
--	@param		workspace	Optional; if a value is passed it will be reused as the client library's workspace. Although a table 
--							is expected, any value but nil is acceptable. Nil or no argument will be replaced with a new table.
--	@return		workspace	The workspace for the new library.
--	@return		interface	The table returned from LibStub. Only returned if 'register' was true.
--	@return		oldMinor	Minor version of the previous instance of the library.  Only returned if 'register' was true.
function core:NewLib(...)
	local register, major, minor, workspace
	local interface, oldMinor		--	only used if 'register' is not false
	
	local t = type(...)
	
	--	shift arguments based on the type of the first
	if t == 'boolean' then
		--	use first value
		register, major, minor, workspace = ...
	elseif t == nil then
		--	replace first value with default
		register, major, minor, workspace = true, select(2, ...)
	else
		--	prepend default
		register, major, minor, workspace = true, ...
	end
	
	--	check for collisions
	if self.clientMajor then
		--	close it properly instead of just 'shortcutting' by overwriting
		local oldMajor, oldMinor = self:CloseLib()
		
		--	We can't use error() because we need to return values for the new lib.
		local message = self:BuildErrorMessage('LEFTOPEN', oldMajor, oldMinor)
		geterrorhandler()(message)
	end
	
	--	check the supplied version info
	if type(major) ~= 'string' or not tonumber(minor) then
		local message = self:BuildErrorMessage('BADVERSION', 'NewLib')
		error(message)
	end
	
	if register then
		interface, oldMinor = LibStub:NewLibrary(major, minor)
	end
	
	--	create the workspace even if LibStub said not 
	--	to create the client library, because the 
	--	client's files need to check.
	
	--	show the frame so we can use OnUpdate
	--	to check if this library was left open.
	self.frame:Show()

	self.clientMajor = major
	self.clientMinor = minor
	
	--	if the client gave us a funky workspace, handling it correctly is up to them
	if workspace == nil then
		workspace = {}
	end
	
	self.clientWorkspace = workspace
	
	return workspace, interface, oldMinor
end


---	Retrieves the version info and workspace for the current client library.
--	@name LibBuilder:GetLib()
--	@return major		Major version of the library.
--	@return minor		Minor version of the library.
--	@return workspace	The workspace for the library.
function core:GetLib()
	if not self.clientMajor then
		local message = self:BuildErrorMessage('NOLIB', 'GetLib')
		error(message)
	end
	
	return self.clientMajor, self.clientMinor, self.clientWorkspace
end


---	Closes the active library. This is used for tracking, 
--	to make sure two libraries don't collide.
--	This function returns the same values as GetLib, as a shortcut 
--	in case the client library needs to do last-minute initialization.
--	@name LibBuilder:CloseLib
--	@see GetLib
function core:CloseLib()
	--	we don't want programmers calling this to 
	--	suppress errors before they open their lib.
	if not self.clientMajor then
		local message = self:BuildErrorMessage('NOLIB', 'CloseLib')
		error(message)
	end
	
	--	the library was closed properly, so disable the OnUpdate frame
	self.frame:Hide()
	
	--	grab values before reset so we can return them
	local major, minor, workspace = self.clientMajor, self.clientMinor, self.clientWorkspace
	self.clientMajor, self.clientMinor, self.clientWorkspace = nil, nil, nil
	
	return major, minor, workspace
end


--	Internal; builds an error message.
--	@param	errorName	Name of the error
--	@param	...			List of arguments applicable to the error type.
--	@return message		The complete error string.
function core:BuildErrorMessage(errorName, ...)
	local errorFormat = self.errorStrings[errorName]
	assert(errorFormat)
	
	local body = errorFormat:format(...)
	local message = self.errorStrings.TEMPLATE:format(body)

	return message
end


--	Internal; used for the core's __call metamethod.
--	If the core table is called with no arguments, it 
--	will call GetLib. If the first argument is a valid key
--	or alias, the associated value will be called with the 
--	remaining  arguments. Otherwise, NewLib will be called 
--	with all arguments.
local function libcall(self, ...)
	local argCount = select('#', ...)
	
	if argCount == 0 then
		return self:GetLib()
	elseif self[...] then
		return self[...](self, select(2, ...))
	elseif self.aliases[...] then
		local targetName = self.aliases[...]
		return self[targetName](self, select(2, ...))
	else
		return self:NewLib(...)
	end
end
setmetatable(core, { __call = libcall })
_G[MAJOR] = core


--	create the shell if one doesn't already exist
local shell = _G[BASE]		--	we'll reuse this local after the if block
if not shell then
	shell = {}
	_G[BASE] = shell
	
	shell.builtBy = VERSION		--	currently unused
	shell.defaultVersion = shellDefaultVersion
	
	--	shell should act like the default version as much as possible, but we 
	--	make no guarantees that new methods won't be inserted into the shell
	
	--	Internal; if one argument is provided and it is a valid index 
	--	in this table, the matching value will be returned. If 
	--	multiple arguments are provided and the first is a valid 
	--	index, the remaining arugments will be passed as a call 
	--	to the matching value. If the first argument is -not- a 
	--	valid index, all arguments will be passed as a call to 
	--	the default library.
	local function shellcall(self, ...)		--	later revision can replace this if it wants to do anything funky with passing self to methods
		if ... == '' or rawget(self, (...)) then	--	don't use normal indexing, in order to send method calls through libcall
			if select('#', ...) == 1 then
				return self[...]
			else
				return self[...](select(2, ...))
			end
		else
			return self[self.defaultVersion](...)
		end
	end
	
	
	--	Internal; returns the default version if key is 
	--	an empty string, otherwise checks for the key in 
	--	the default library.  Note that lua will not call 
	--	this metamethod if the specified key is present 
	--	in the shell.
	local function shellindex(self, key)
		--	FIXME: apply aliases to shell?
		if key == nil then
			return nil
		end
		
		local defaultLib = rawget(self, self.defaultVersion)
		
		if key == '' then
			return defaultLib		--	may be nil
		else
			if defaultLib == nil then
				--	can't go further if default isn't present
				local message = core:BuildErrorMessage('SHELLNODEFAULT')
				error(message)
			end
			
			--	we'll get around the 'wrong self' issue by overloading 
			--	__newindex to point back to defaultLib
			return defaultLib[key]
		end
	end
	
	
	local function shellnewindex(self, key, value)
		--	new versions of the library should be set with rawset, 
		--	to bypass this function
		local defaultLib = self[self.defaultVersion]
		
		defaultLib[key] = value
	end
	
	
	setmetatable(shell, { __call = shellcall, __index = shellindex, __newindex = shellnewindex})
end

--	add current version to the shell
rawset(shell, VERSION, core)
